@@ -1,7 +1,7 @@
 /* -------------------------------------------------------------------------
  * pg_dumplo
  *
- * $Header: /cvsroot/pgsql-server/contrib/pg_dumplo/utils.c,v 1.6 2002/09/05 21:01:16 tgl Exp $
+ * $PostgreSQL: pgsql/contrib/pg_dumplo/utils.c,v 1.9 2004/08/29 05:06:36 momjian Exp $
  *
  *					Karel Zak 1999-2000
  * -------------------------------------------------------------------------
@@ -30,6 +30,7 @@ void
 index_file(LODumpMaster * pgLO)
 {
 	char		path[BUFSIZ];
+	int			sz;
 
 	if (pgLO->action == ACTION_SHOW)
 		return;
@@ -49,7 +50,8 @@ index_file(LODumpMaster * pgLO)
 			}
 		}
 
-		snprintf(path, BUFSIZ, "%s/lo_dump.index", path);
+		sz = strlen(path);
+		strncat(path, "/lo_dump.index", BUFSIZ - sz);
 
 		if ((pgLO->index = fopen(path, "w")) == NULL)
 		{
@@ -60,8 +62,8 @@ index_file(LODumpMaster * pgLO)
 	}
 	else if (pgLO->action != ACTION_NONE)
 	{
-
-		snprintf(path, BUFSIZ, "%s/lo_dump.index", path);
+		sz = strlen(path);
+		strncat(path, "/lo_dump.index", BUFSIZ - sz);
 
 		if ((pgLO->index = fopen(path, "r")) == NULL)
 		{
